#include <windows.h>
#include <math.h>
#include <stdio.h>
#include "util.h"
#include "mywindow.h"
#include "fishEye.h"
#include "errHandle.h"

FishEye myFishEye;
PeekHole myPeekHole;

const int RgnOffset=5;

void FishEye::init(HDC hdc, POINT myCoord, int myRadius, float myDensity)
{
	radius=myRadius;
	internalRadius=myDensity*radius;
	width=radius*PI/2.0;
	if (hMemDC!=NULL)
	{
		DeleteDC(hMemDC);
		DeleteObject(hBitmap);	
		DeleteObject(hRgn);
	}

	hMemDC=CreateCompatibleDC(hdc);
	hBitmap=CreateCompatibleBitmap(hdc, width*2, width*2);
	SelectObject(hMemDC, hBitmap);

	coord.x=myCoord.x;
	coord.y=myCoord.y;

	BitBlt(hMemDC, 0, 0, 2*width, 2*width, hdc, coord.x-width, coord.y-width, 
			SRCCOPY);
	hRgn=CreateEllipticRgn(coord.x-radius-RgnOffset, coord.y-radius-RgnOffset, 
		coord.x+radius+RgnOffset, coord.y+radius+RgnOffset);

}
	

void FishEye::drawFishEye(HDC hdc, POINT myCoord, bool erase)
{
	float dist, ratio;
	int result;
	char msg[128];
	HWND hwnd;

	if (erase)
	{
		result=ExtSelectClipRgn(hdc, hRgn, RGN_COPY);
		if (result==ERROR)
		{			
			errHandle("clipping");
		}
		if (result==NULLREGION)
		{
			errHandle("null");
		}
		BitBlt(hdc, coord.x-radius, coord.y-radius, 2*radius, 2*radius, hMemDC, 
			width-radius, width-radius,	SRCCOPY);

		result=ExtSelectClipRgn(hdc, NULL, RGN_COPY);
		
	}
	DeleteObject(hRgn);
	coord.x=myCoord.x;
	coord.y=myCoord.y;
	hRgn=CreateEllipticRgn(coord.x-radius-RgnOffset, coord.y-radius-RgnOffset, 
		coord.x+radius+RgnOffset, coord.y+radius+RgnOffset);
	BitBlt(hMemDC, 0, 0, 2*width, 2*width, hdc, coord.x-width, coord.y-width, 
			SRCCOPY);
	for (int row=0; row<radius; row++)
	{
		for (int col=0; col<radius; col++)
		{
			dist=sqrtf(row*row+col*col);
			if (dist>=radius)
			{
				continue;
			}
			ratio=(float)radius*asin(dist/(float)radius)/dist;
			
			drawPixels(hdc, ratio, row, col);
		}
	}
}

void FishEye::drawPixels(HDC hdc, float ratio, int row, int col)
{
	POINT point;
	COLORREF color;
	int offsetX, offsetY, x, y;
	offsetX=col*ratio;
	offsetY=row*ratio;

	x=width+offsetX;
	y=width+offsetY;

	if (x<2*width&&y<2*width&&x>=0&&y>=0)
	{
		color=GetPixel(hMemDC, x, y);
		SetPixel(hdc, coord.x+col, coord.y+row, color);
	}

	x=width+offsetX;
	y=width-offsetY;
	if (x<2*width&&y<2*width&&x>=0&&y>=0)
	{
		color=GetPixel(hMemDC, x, y);
		SetPixel(hdc, coord.x+col, coord.y-row, color);
	}

	x=width-offsetX;
	y=width+offsetY;
	if (x<2*width&&y<2*width&&x>=0&&y>=0)
	{
		color=GetPixel(hMemDC, x, y);
		SetPixel(hdc, coord.x-col, coord.y+row, color);
	}
	x=width-offsetX;
	y=width-offsetY;
	if (x<2*width&&y<2*width&&x>=0&&y>=0)
	{
		color=GetPixel(hMemDC, x, y);
		SetPixel(hdc, coord.x-col, coord.y-row, color);
	}
}


void PeekHole::drawPixels(HDC hdc, float ratio, POINT coord, int row, int col, RECT& rect)
{
	POINT point;
	COLORREF color;
	int offsetX, offsetY, x, y;
	offsetX=col*ratio;
	offsetY=row*ratio;
	
	x=coord.x+offsetX;
	y=coord.y+offsetY;

	if (x>rect.left&&x<rect.right&&y>rect.top&&y<rect.bottom)
	{
		color=GetPixel(hdc, x, y);
		SetPixel(hdc, holeCoord.x+col, holeCoord.y+row, color);
	}


	x=coord.x+offsetX;
	y=coord.y-offsetY;
	if (x>rect.left&&x<rect.right&&y>rect.top&&y<rect.bottom)
	{
		color=GetPixel(hdc, x, y);
		SetPixel(hdc, holeCoord.x+col, holeCoord.y-row, color);
	}
	

	x=coord.x-offsetX;
	y=coord.y+offsetY;
	if (x>rect.left&&x<rect.right&&y>rect.top&&y<rect.bottom)
	{
		color=GetPixel(hdc, x, y);
		SetPixel(hdc, holeCoord.x-col, holeCoord.y+row, color);
	}

	x=coord.x-offsetX;
	y=coord.y-offsetY;
	if (x>rect.left&&x<rect.right&&y>rect.top&&y<rect.bottom)
	{
		color=GetPixel(hdc, x, y);
		SetPixel(hdc, holeCoord.x-col, holeCoord.y-row, color);
	}
		
}

void PeekHole::init(POINT myHoleCoord, int myHoleRadius, int myViewRadius)
{
	memcpy(&holeCoord, &myHoleCoord, sizeof(POINT));
	holeRadius=myHoleRadius;
	viewRadius=myViewRadius;
}


void PeekHole::peek(HWND hwnd, POINT myViewCoord)
{
	float dist, ratio;
	RECT rect;
	HDC hdc;
	POINT origin;
	clientAreaOrg(origin, hwnd);
	
	hdc=GetDC(hwnd);
	GetClientRect(hwnd, &rect);
	rect.left+=origin.x;
	rect.top+=origin.y;
	rect.right-=origin.x;
	rect.bottom-=origin.y;
	for (int row=0; row<holeRadius; row++)
	{
		for (int col=0; col<holeRadius; col++)
		{
			dist=sqrtf(row*row+col*col);
			if (dist>=holeRadius||dist==0)
			{
				continue;
			}
			ratio=(float)viewRadius*asin(dist/(float)holeRadius)/dist;
			
			drawPixels(hdc, ratio, myViewCoord, row, col, rect);
		}
	}
	ReleaseDC(hwnd, hdc);
}


long CALLBACK fishEye(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	static bool startDrawing=false;
	static POINT coord;
	HDC hdc;
	switch (message)
	{
	case WM_CREATE:
		break;
	case WM_LBUTTONDOWN:
		hdc=GetDC(hwnd);
		coord.x=LOWORD(lParam);
		coord.y=HIWORD(lParam);
		if (!startDrawing)
		{
			myPeekHole.init(coord, 100, 200);
			startDrawing=true;	
		}
		else
		{
			myPeekHole.peek(hwnd, coord);
		}
		ReleaseDC(hwnd, hdc);
		break;
	}

	return 0L;
}


