#include <windows.h>
#include <vector>
#include <set>
#include "errHandle.h"
#include <windows.h>
#include "region.h"
#include "util.h"
#include "mywindow.h"
#include "mywindowProc.h"






/*
enum Position
{
	NW=0, N=1, NE=2, E=3, SE=4, S=5, SW=6, W=7
};
*/
bool *pHistory=NULL;

const unsigned char RedTolerance=0;
const unsigned char GreenTolerance=0;
const unsigned char BlueTolerance=0;


bool compareColor(RGBTRIPLE* src, RGBTRIPLE* dest)
{
	return abs(src->rgbtRed-dest->rgbtRed)<=RedTolerance
		&&abs(src->rgbtGreen- dest->rgbtGreen)<=GreenTolerance
		&&abs(src->rgbtBlue- dest->rgbtBlue)<=BlueTolerance;
}


bool isSameColor(POINT src, POINT dest, int rowBytes, int channel, uchar* pImage)
{
	RGBTRIPLE* pSrc, *pDest;

	pSrc=(RGBTRIPLE*)(pImage+src.y*rowBytes+src.x*channel);
	pDest=(RGBTRIPLE*)(pImage+dest.y*rowBytes+dest.x*channel);
	return compareColor(pSrc, pDest);
	//return memcmp(pSrc, pDest, 3)==0;
}


void gridOp(long& x, long& y, int dir)
{
	switch (dir)
	{
	case NW:
		x--;
		y--;
		break;
	case N:
		y--;
		break;
	case NE:
		x++;
		y--;
		break;
	case E:
		x++;
		break;
	case SE:
		x++;
		y++;
		break;
	case S:
		y++;
		break;
	case SW:
		x--;
		y++;
		break;
	case W:
		x--;
		break;
	}
}

bool getCoord(POINT coord, POINT& result, int width, int height, int dir)
{
	result.x=coord.x;
	result.y=coord.y;
	/*
	switch (pos)
	{
	case NW:
		result.x=coord.x-1;
		result.y=coord.y-1;
		break;
	case N:
		result.x=coord.x;
		result.y=coord.y-1;
		break;
	case NE:
		result.x=coord.x+1;
		result.y=coord.y-1;
		break;
	case E:
		result.x=coord.x+1;
		result.y=coord.y;
		break;
	case SE:
		result.x=coord.x+1;
		result.y=coord.y+1;
		break;
	case S:
		result.x=coord.x;
		result.y=coord.y+1;
		break;
	case SW:
		result.x=coord.x-1;
		result.y=coord.y+1;
		break;
	case W:
		result.x=coord.x-1;
		result.y=coord.y;
		break;
	}
	*/
	gridOp(result.x, result.y, dir);
	return (result.x>=0&&result.y>=0&&result.x<width&&result.y<height);
}


bool isBoundary(POINT coord, int width, int height, int rowBytes, int channel, 
				uchar* pImage)
{
	POINT point;
	bool hasSame=false;
	bool hasDifferent=false;
	for (int i=0; i<8; i++)
	{
		if (getCoord(coord, point, width, height, i))
		{
			if (!isSameColor(coord, point, rowBytes, channel, pImage))
			{
				hasDifferent=true;
			}
			else
			{
				hasSame=true;
			}
		}	
		else
		{
			return true;
		}
	}
	return hasSame&&hasDifferent;
}


bool isInternal(POINT coord, int width, int height, int rowBytes, int channel, uchar* pImage)
{
	POINT point;
	for (int i=1; i<8; i+=2)
	{
		if (getCoord(coord, point, width, height, i))
		{
			if (!isSameColor(coord, point, rowBytes, channel, pImage))
			{
				return false;
			}
		}
		else
		{
			return false;
		}
	}
	return true;
}

bool samePoint(const POINT& p1, const POINT& p2)
{
	return p1.x==p2.x && p1.y==p2.y;
}

bool isPointInPath(POINT coord, PointVector& path)
{
	for (int i=0; i<path.size(); i++)
	{
		if (samePoint(path[i], coord))
		{
			return true;
		}
	}
	return false;
}

bool hasInternalNeighbour(POINT coord, BITMAP& bitmap, uchar* pImage)
{
	POINT point;
	for (int i=0; i<8; i++)
	{
		if (getCoord(coord, point, bitmap.bmWidth, bitmap.bmHeight, i))
		{
			if (isInternal(point, bitmap.bmWidth, bitmap.bmHeight, bitmap.bmWidthBytes,
				bitmap.bmBitsPixel/8, pImage))
			{
				return true;
			}
		}
	}
	return false;
}

bool isPointInHistory(POINT coord, PointSet& history)
{
	return true;
}

bool doCreateClosedPath(POINT coord, BITMAP& bitmap, uchar* pImage,
						PointVector& path, PointSet& history)
{
	int i, counter=0;
	POINT point;
	int index[8];
	bool deadEnd=true;

	path.push_back(coord);
	history.insert(coord);
	for (i=0; i<8; i++)
	{
		if (getCoord(coord, point, bitmap.bmWidth, bitmap.bmHeight, i))
		{
			if (isSameColor(coord, point, bitmap.bmWidthBytes, bitmap.bmBitsPixel/8, pImage))
			{
				if (!isInternal(point, bitmap.bmWidth, bitmap.bmHeight, bitmap.bmWidthBytes, 
					bitmap.bmBitsPixel/8, pImage))
				{
					if (hasInternalNeighbour(point, bitmap, pImage))
					{
						if (!isPointInPath(point, path)&&history.find(point)==history.end())
						{
							doCreateClosedPath(point, bitmap, pImage, path, history);
							deadEnd=false;
						}
						else
						{
							index[counter]=i;
							counter++;
						}						
					}					
				}
			}
		}		
	}
	if (!deadEnd)
	{
		return true;
	}
	//else
	if (counter>0)
	{
		//it must meet our previous path!
		for (i=0; i<counter; i++)
		{		
			if (getCoord(coord, point, bitmap.bmWidth, bitmap.bmHeight, index[i]))
			{
				if (samePoint(path[0], point))
				{
					path.push_back(point);
					return true;
				}
			}
		}
	}

	//dead end but counter==0 which means no meeting with previous path
	//errHandle("dead end but not meet any path! error????\n");
	return false;	
}


HRGN createRgnFromPath(HDC hdc, PointVector& path)
{
	BeginPath(hdc);
	MoveToEx(hdc, path[0].x, path[0].y, NULL);
	for (int i=1; i<path.size(); i++)
	{
		LineTo(hdc, path[i].x, path[i].y);
	}
	EndPath(hdc);
	return PathToRegion(hdc);
}




HRGN createRgn(HDC hdc, POINT coord, BITMAP& bitmap, uchar* pImage, PointVector& path, 
			   PointSet& history)
{
	bool result=false;

	if (history.find(coord)!=history.end())
	{
		return NULL;
	}

	if (isInternal(coord, bitmap.bmWidth, bitmap.bmHeight, bitmap.bmWidthBytes, 
		bitmap.bmBitsPixel/8, pImage))
	{
		return NULL;
	}
	path.clear();
	//BeginPath(hdc);
	if (doCreateClosedPath(coord, bitmap, pImage, path, history))
	{
		if (path.size()>=4)
		{
			return createRgnFromPath(hdc, path);
		}
	}
	return NULL;
	//return CreatePolygonRgn(&(*(path.begin())), path.size(), WINDING);
}

RGBTRIPLE* getColor(BITMAP& bitmap, uchar* pImage, int x, int y)
{
	return (RGBTRIPLE*)(pImage+y*bitmap.bmWidthBytes+x*(bitmap.bmBitsPixel/8));
}
	
//we can expand our method to allow noise
bool similarColor(RGBTRIPLE* first, RGBTRIPLE* second)
{
	return memcmp(first, second, sizeof(RGBTRIPLE))==0;
}


	
void drawRegionByIndex(HDC hdc, HRGN hRgn, int index)
{
	PALETTEENTRY palEntry;
	HBRUSH hBrush=NULL; 
	unsigned char r, g, b;
	r=(index%8)*32;
	index/=8;
	g=(index%8)*32;
	index/=8;
	b==(index%8)*32;

	hBrush=CreateSolidBrush(RGB(r, g, b));
	int result=FillRgn(hdc, hRgn, hBrush);
	//FrameRgn(hdc, hRgn, hBrush, 3, 3);
	DeleteObject(hBrush);
}


/*
void drawRegionByIndex(HDC hdc, HRGN hRgn, int index)
{
	PALETTEENTRY palEntry;
	HBRUSH hBrush=NULL; 
	RECT rect;

	GetSystemPaletteEntries(hdc, index%9+1, 1, &palEntry);

	hBrush=CreateSolidBrush(RGB(palEntry.peRed, palEntry.peGreen, palEntry.peBlue));
	GetRgnBox(hRgn, &rect);
	int result=FillRgn(hdc, hRgn, hBrush);
	GetRgnBox(hRgn, &rect);
	DeleteObject(hBrush);
}
*/



int getRegionData(HRGN hRgn, RGNDATA*& pRgnData)
{
	int iSize;
	unsigned char* buf;
	RGNDATAHEADER* pRgnDataHeader;

	iSize=GetRegionData(hRgn, 0, NULL);
	if (iSize==0)
	{
		return -1;
	}
	buf=new unsigned char[iSize];
	pRgnDataHeader=(RGNDATAHEADER*)buf;
	pRgnDataHeader->dwSize=sizeof(RGNDATAHEADER);
	pRgnDataHeader->iType=RDH_RECTANGLES;
	if (GetRegionData(hRgn, iSize, (LPRGNDATA)buf)==0)
	{
		return -1;
	}
	pRgnData=(LPRGNDATA)buf;
	return iSize;
}


int getRegionArea(HRGN hRgn)
{
	PRGNDATA pRgnData;
	RECT* pRect;
	int iSize;
	int result=0;
	iSize=getRegionData(hRgn, pRgnData);

	if (iSize==-1)
	{
		return -1;
	}
	
	pRect=(RECT*)pRgnData->Buffer;
	for (int i=0; i<pRgnData->rdh.nCount; i++)
	{
		result+=(pRect->bottom-pRect->top)*(pRect->right-pRect->left);
		pRect++;
	}
	delete pRgnData;
	return result;
}
		



HRGN transformRegion(HRGN hRgn, const XFORM* xForm)
{
	HRGN result;
	PRGNDATA pRgnData;
	unsigned int iSize;
	iSize=getRegionData(hRgn, pRgnData);
	result= ExtCreateRegion(xForm, iSize, pRgnData);
	delete pRgnData;
	return result;
}





void displayRgn(HDC hdc, HRGN hRgn)
{
	PALETTEENTRY palEntry;
	HBRUSH hBrush=NULL; 

	PRGNDATA pRgnData;
	RECT* pRect;
	
	getRegionData(hRgn, pRgnData);

	pRect=(RECT*)(pRgnData->Buffer);
	for (int i=0; i<pRgnData->rdh.nCount; i++)
	{		
		palEntry.peFlags=NULL;
		GetSystemPaletteEntries(hdc, i%9+1, 1, &palEntry);
			
		hBrush=CreateSolidBrush(RGB(palEntry.peRed, palEntry.peGreen, palEntry.peBlue));
		FillRect(hdc, pRect, hBrush);
		DeleteObject(hBrush);
		pRect++;
	}
	delete pRgnData;
}

bool pointIncluded(RgnVector& rgnVector, int x, int y)
{
	for (int i=0; i<rgnVector.size(); i++)
	{
		if (PtInRegion(rgnVector[i], x, y))
		{
			return true;
		}
	}
	return false;
}

int findFirstDifferent(POINT coord, int startPos, BITMAP& bitmap, uchar* pImage)
{
	POINT point;
	for (int i=0; i<8; i++)
	{
		if (getCoord(coord, point, bitmap.bmWidth, bitmap.bmHeight, 
			i+startPos))
		{
			if (!isSameColor(coord, point, bitmap.bmWidthBytes, bitmap.bmBitsPixel/8, 
				pImage))
			{
				return (i+startPos)%8;
			}
		}
	}
	errHandle("impossible");
	return -1;
}

bool isSamePoint(const POINT& first, const POINT& second)
{
	return first.x==second.x&&first.y==second.y;
}

int findFirstSame(POINT coord, BITMAP& bitmap, uchar* pImage, int startPos)
{
	POINT point;
	for (int i=0; i<8; i++)
	{
		if (getCoord(coord, point, bitmap.bmWidth, bitmap.bmHeight, 
			i+startPos))
		{
			if (isSameColor(coord, point, bitmap.bmWidthBytes, bitmap.bmBitsPixel/8, 
				pImage))
			{
				return (i+startPos)%8;
			}
		}
	}
	errHandle("impossible, isolated");
	return -1;
}

/*
int findFirstBoundary(POINT coord, BITMAP& bitmap, uchar* pImage, int dir)
{
	POINT point;
	for (int i=0; i<8; i++)
	{
		if (getCoord(coord, point, bitmap.bmWidth, bitmap.bmHeight, (dir+i)%8))
		{
			if (isBoundary(point, bitmap.bmWidth, bitmap.bmHeight, bitmap.bmWidthBytes,
				bitmap.bmBitsPixel/8, pImage))
			{
				return (dir+i)%8;
			}
		}
	}
	return -1;
}
*/



int findFirstSameColorBoundary(POINT coord, BITMAP& bitmap, uchar* pImage, int dir)
{
	POINT point;
	for (int i=0; i<8; i++)
	{
		if (getCoord(coord, point, bitmap.bmWidth, bitmap.bmHeight, (dir+i)%8))
		{
			if (isSameColor(coord, point, bitmap.bmWidthBytes, bitmap.bmBitsPixel/8,
				pImage))
			{
				if (isBoundary(point, bitmap.bmWidth, bitmap.bmHeight, 
					bitmap.bmWidthBytes, bitmap.bmBitsPixel/8, pImage))
				{
					return (dir+i)%8;
				}
			}
		}
	}
	return -1;
}

bool isLeftBoundary(POINT coord, int width, int height, int rowBytes, int channel, 
				uchar* pImage)
{
	POINT point;
	bool hasNeighbour=false;
	bool isLeftEmpty=false;//it 

	//
	if (getCoord(coord, point, width, height, 7))
	{
		if (isSameColor(coord, point, rowBytes, channel, pImage))
		{
			return false;
		}
	}
	//it must be a boundary case, at edge or has neighbour of different color

	//now we don't want it to be singleton
	for (int i=0; i<7; i++)
	{
		if (getCoord(coord, point, width, height, i))
		{
			if (isSameColor(coord, point, rowBytes, channel, pImage))
			{
				return true;
			}
		}	
	}
	return false;
}



void doScanRegion(HDC hdc, POINT coord, BITMAP& bitmap, uchar* pImage, PointVector& path)
{
	int startPos=0;

	POINT point, result;
	if ((startPos=findFirstSameColorBoundary(coord, bitmap, pImage, startPos))!=-1)
	{
		getCoord(coord, point, bitmap.bmWidth, bitmap.bmHeight, startPos);
		pHistory[coord.y*bitmap.bmWidth+coord.x]=true;
		result.x=coord.x;
		result.y=bitmap.bmHeight-coord.y-1;
		path.push_back(result);
		MoveToEx(hdc, result.x, result.y, NULL);
		do
		{
			pHistory[point.y*bitmap.bmWidth+point.x]=true;
			result.x=point.x;
			result.y=bitmap.bmHeight-point.y-1;
			path.push_back(result);
			LineTo(hdc, result.x, result.y);

			if (startPos%2==1)
			{
				startPos=(startPos+7)%8;
			}
			else
			{
				startPos=(startPos+6)%8;
			}
		
			if ((startPos=findFirstSameColorBoundary(point, bitmap, 
				pImage, startPos))!=-1)
			{
				getCoord(point, point, bitmap.bmWidth, bitmap.bmHeight,
					startPos);
			}
		}while (!isSamePoint(coord, point));
		result.x=coord.x;
		result.y=bitmap.bmHeight-coord.y-1;
		path.push_back(result);
	}	
}




void createRegionFromPath(HDC hdc, PointVector& path, RgnVector& rgnVector)
{
	if (path.size()<4)
	{
		return;
	}
	BeginPath(hdc);
	MoveToEx(hdc, path[0].x, path[0].y, NULL);
	for (int i=1; i<path.size(); i++)
	{
		LineTo(hdc, path[i].x, path[i].y);
	}
	EndPath(hdc);
	rgnVector.push_back(PathToRegion(hdc));
}

void outputPoint(PointVector& path, int index)
{
	char fileName[32];
	char buf[32];
	FILE* stream;
	sprintf(fileName, "%d-%d.txt", index, path.size());
	stream=fopen(fileName, "w");
	for (int i=0; i<path.size(); i++)
	{
		sprintf(buf, "{%d,%d},", path[i].x, path[i].y);
		fwrite(buf, strlen(buf), 1, stream);
	}
	fclose(stream);
}


void createRegionFromPath2(HDC hdc, PointVector& path, RgnVector& rgnVector)
{
	if (path.size()<4)
	{
		return;
	}
	rgnVector.push_back(CreatePolygonRgn(&(*path.begin()), path.size(), WINDING));

	//outputPoint(path, rgnVector.size());
}

void scanRegions(HDC hdc, BITMAP& bitmap, uchar* pImage, RgnVector& rgnVector)
{
	POINT coord;
	PointVector path;
	HPEN hPen;
	hPen=CreatePen(PS_SOLID, 1, RGB(255,0,0));
	hPen=(HPEN)SelectObject(hdc, hPen);
	
	pHistory=new bool[bitmap.bmWidth*bitmap.bmHeight];
	if (pHistory==NULL)
	{
		errHandle("unable allocate memory");
	}
	memset(pHistory, false, bitmap.bmWidth*bitmap.bmHeight);
	for (int y=0; y<bitmap.bmHeight; y++)
	{
		for (int x=0; x<bitmap.bmWidth; x++)
		{			
			//if (!pointIncluded(rgnVector, x, bitmap.bmHeight-y-1)&&
			//	!pHistory[y*bitmap.bmWidth+x])
			if (!pHistory[y*bitmap.bmWidth+x])
			{
				coord.x=x;
				coord.y=y;
				if (isLeftBoundary(coord, bitmap.bmWidth, 
					bitmap.bmHeight, bitmap.bmWidthBytes, 
					bitmap.bmBitsPixel/8, pImage))
				{
					path.clear();
					doScanRegion(hdc, coord, bitmap, pImage, path);
					createRegionFromPath2(hdc, path, rgnVector);
				}
			}
			pHistory[y*bitmap.bmWidth+x]=true;
		}
	}
	hPen=(HPEN)SelectObject(hdc, hPen);
	DeleteObject(hPen);
	delete[]pHistory;
}





long CALLBACK myRegion(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	int i;
	static int counter=0;
	HBITMAP hBitmap;
	static POINT coord;
	HDC hdc, hMemDC;
	BITMAP bitmap;
	BITMAPINFO bmInfo;
	static RgnVector rgnVector;
	unsigned char* pImage;
	RECT rect;
	char msg[128];
	HRGN hRgn, hRectRgn;
	HBRUSH hBrush, hRedBrush, hBlueBrush;
	HPEN hRedPen, hGreenPen;
	XFORM xForm;
	static TCHAR pstrFileName[MAX_PATH], pstrTitleName[MAX_PATH];
	switch (message)
	{
	case WM_CREATE:
		break;
	case WM_LBUTTONDOWN:
		hdc=GetDC(hwnd);
		hMemDC=CreateCompatibleDC(hdc);
		coord.x=LOWORD(lParam);
		coord.y=HIWORD(lParam);
		if (!FileOpenDlg(hwnd, pstrFileName, pstrTitleName))
		{
			return 0;
		}

		if ((hBitmap=(HBITMAP)LoadImage(NULL, pstrFileName, IMAGE_BITMAP, 0, 0,
			LR_LOADFROMFILE))==NULL)
		{
			return 0L;
		}
		SelectObject(hMemDC, hBitmap);
		GetObject(hBitmap, sizeof(BITMAP), &bitmap);
		pImage=new unsigned char[bitmap.bmWidthBytes*bitmap.bmHeight];
		init(bitmap, bmInfo.bmiHeader);
		GetDIBits(hdc, hBitmap, 0, bitmap.bmHeight, pImage, &bmInfo, DIB_RGB_COLORS);
		//analysis(hdc, bitmap, pImage, rgnVector);
		scanRegions(hdc, bitmap, pImage, rgnVector);


		//StretchBlt(hdc, coord.x, coord.y, bitmap.bmWidth*1, bitmap.bmHeight*1, hMemDC, 0, 0, 
		//	bitmap.bmWidth, bitmap.bmHeight, SRCCOPY);
		
		sprintf(msg, "total %d", rgnVector.size());
		SetWindowText(hwnd, msg);
		DeleteObject(hBitmap);
		DeleteDC(hMemDC);
		ReleaseDC(hwnd, hdc);
		break;
	case WM_RBUTTONDOWN:

		if (counter<rgnVector.size())
		{					
			hdc=GetDC(hwnd);
			drawRegionByIndex(hdc, rgnVector[counter], counter);
			counter++;
			ReleaseDC(hwnd, hdc);
		}
		else
		{
			for (i=0; i<rgnVector.size(); i++)
			{
				DeleteObject(rgnVector[i]);
			}
			rgnVector.clear();
		}

	}
	return 0L;
}
			





