#include <windows.h>
#include <math.h>
#include <stdio.h>
#include "util.h"
#include "myWindow.h"
#include "errHandle.h"

void rotate(float degree, XFORM& xForm)
{
	float sine, cose;
	float radi;
	radi=degree/180.0*3.14159;
	sine=sinf(radi);
	cose=cosf(radi);
	xForm.eM11 = cose; 
    xForm.eM12 = sine; 
    xForm.eM21 = -sine; 
    xForm.eM22 = cose; 
    xForm.eDx  =  0.0; 
    xForm.eDy  =  0.0; 
}

void translate(float x, float y, XFORM& xForm)
{
	xForm.eM11 = 1; 
    xForm.eM12 = 0; 
    xForm.eM21 = 0; 
    xForm.eM22 = 1; 
    xForm.eDx  =  x; 
    xForm.eDy  =  y; 
}

void shear(float hor, float ver, XFORM& xForm)
{
	xForm.eM11 = 1; 
    xForm.eM12 = hor; 
    xForm.eM21 = ver; 
    xForm.eM22 = 1; 
    xForm.eDx  = 0; 
    xForm.eDy  = 0;
}

void reflect(float hor, float ver, XFORM& xForm)
{
	xForm.eM11 = hor; 
    xForm.eM12 = 0; 
    xForm.eM21 = 0; 
    xForm.eM22 = ver; 
    xForm.eDx  =  0; 
    xForm.eDy  =  0;
}

void scale(float hor, float ver, XFORM& xForm)
{
	xForm.eM11 = hor; 
    xForm.eM12 = 0; 
    xForm.eM21 = 0; 
    xForm.eM22 = ver; 
    xForm.eDx  =  0; 
    xForm.eDy  =  0;
}


void identity(XFORM& xForm)
{
	xForm.eM11 = 1; 
    xForm.eM12 = 0; 
    xForm.eM21 = 0; 
    xForm.eM22 = 1; 
    xForm.eDx  =  0; 
    xForm.eDy  =  0;
}


// GetRotatedBitmapNT	- Create a new bitmap with rotated image
// Returns		- Returns new bitmap with rotated image
// hBitmap		- Bitmap to rotate
// radians		- Angle of rotation in radians
// clrBack		- Color of pixels in the resulting bitmap that do
//			  not get covered by source pixels
HBITMAP GetRotatedBitmap(HDC hdc, HBITMAP hBitmap, float radians, COLORREF clrBack )
{
	// Create a memory DC compatible with the display
	HDC sourceDC, destDC;
	// Get logical coordinates
	BITMAP bm;
	HBITMAP hOldSrcBitmap, hOldDestBitmap;
	float cosine = (float)cos(radians);
	float sine = (float)sin(radians);

	sourceDC=CreateCompatibleDC( hdc );
	destDC=CreateCompatibleDC( hdc );


	GetObject( hBitmap, sizeof( bm ), &bm );

	// Compute dimensions of the resulting bitmap
	// First get the coordinates of the 3 corners other than origin
	int x1 = (int)(bm.bmHeight * sine);
	int y1 = (int)(bm.bmHeight * cosine);
	int x2 = (int)(bm.bmWidth * cosine + bm.bmHeight * sine);
	int y2 = (int)(bm.bmHeight * cosine - bm.bmWidth * sine);
	int x3 = (int)(bm.bmWidth * cosine);
	int y3 = (int)(-bm.bmWidth * sine);

	int minx = min(0,min(x1, min(x2,x3)));
	int miny = min(0,min(y1, min(y2,y3)));
	int maxx = max(0,max(x1, max(x2,x3)));
	int maxy = max(0,max(y1, max(y2,y3)));

	int w = maxx - minx;
	int h = maxy - miny;

	// Create a bitmap to hold the result
	HBITMAP hbmResult = CreateCompatibleBitmap(hdc, w, h);

	hOldSrcBitmap=(HBITMAP)SelectObject( sourceDC, hBitmap );
	hOldDestBitmap=(HBITMAP)SelectObject( destDC, hbmResult );

	// Draw the background color before we change mapping mode
	HBRUSH hbrBack = CreateSolidBrush( clrBack );
	SelectObject( destDC, hbrBack );
	PatBlt(destDC, 0, 0, w, h, PATCOPY );
	SelectObject(destDC, (HBRUSH)GetStockObject(WHITE_BRUSH));
	DeleteObject( hbrBack );

	// We will use world transform to rotate the bitmap
	SetGraphicsMode(destDC, GM_ADVANCED);
	XFORM xform;
	xform.eM11 = cosine;
	xform.eM12 = -sine;
	xform.eM21 = sine;
	xform.eM22 = cosine;
	xform.eDx = (float)-minx;
	xform.eDy = (float)-miny;

	SetWorldTransform( destDC, &xform );

	// Now do the actual rotating - a pixel at a time
	int result=BitBlt(destDC, 0,0,bm.bmWidth, bm.bmHeight, sourceDC, 0, 0, SRCCOPY );

	SelectObject(destDC, hOldDestBitmap);
	SelectObject(sourceDC, hOldSrcBitmap);
	DeleteDC(destDC);
	DeleteDC(sourceDC);


	return hbmResult;
}




