#include <windows.h>
#include <winuser.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "myWindow.h"
#include "transform.h"
#include "util.h"
#include "errHandle.h"

#define EVEN(x) (x%2==0)
#define ODD(x)  (x%2==1)

extern char buf[256];

OPENFILENAME ofn;
static TCHAR szBmpFilter[] = TEXT ("PNG Files (*.bmp)\0*.BMP\0")
        TEXT ("All Files (*.*)\0*.*\0\0");

const unsigned char PNG_Signature[8]={137, 80, 78, 71, 13, 10, 26, 10};

unsigned long byteSwap (unsigned long nLongNumber)
{
   union u {unsigned long vi; unsigned char c[sizeof(unsigned long)];}; 
   union v {unsigned long ni; unsigned char d[sizeof(unsigned long)];};
   union u un; 
   union v vn; 
   un.vi = nLongNumber; 
   vn.d[0]=un.c[3]; 
   vn.d[1]=un.c[2]; 
   vn.d[2]=un.c[1]; 
   vn.d[3]=un.c[0]; 
   return (vn.ni); 
}

unsigned long int readChunk(FILE* file, unsigned char typeBuf[4])
{
	unsigned int size;
	if (fread(&size, 4, 1, file)!=1)
	{
		return -1;
	}
	size=byteSwap(size);
	if (fread(typeBuf, 4, 1, file)!=1)
	{
		return -1;
	}


	return size;
}

void init(const BITMAP& bitmap, BITMAPINFOHEADER& bmHeader)
{
	memset(&bmHeader, 0, sizeof(BITMAPINFOHEADER));
	bmHeader.biSize=sizeof(BITMAPINFOHEADER);
	bmHeader.biWidth=bitmap.bmWidth;
	bmHeader.biHeight=bitmap.bmHeight;
	bmHeader.biCompression=BI_RGB;
	bmHeader.biPlanes=1;
	bmHeader.biBitCount=bitmap.bmBitsPixel;
}

void FileInitialize (HWND hwnd)
{
    ofn.lStructSize       = sizeof (OPENFILENAME);
    ofn.hwndOwner         = hwnd;
    ofn.hInstance         = NULL;
    ofn.lpstrFilter       = szBmpFilter;
    ofn.lpstrCustomFilter = NULL;
    ofn.nMaxCustFilter    = 0;
    ofn.nFilterIndex      = 0;
    ofn.lpstrFile         = NULL;          // Set in Open and Close functions
    ofn.nMaxFile          = MAX_PATH;
    ofn.lpstrFileTitle    = NULL;          // Set in Open and Close functions
    ofn.nMaxFileTitle     = MAX_PATH;
    ofn.lpstrInitialDir   = NULL;
    ofn.lpstrTitle        = NULL;
    ofn.Flags             = 0;             // Set in Open and Close functions
    ofn.nFileOffset       = 0;
    ofn.nFileExtension    = 0;
    ofn.lpstrDefExt       = TEXT ("*.*");
    ofn.lCustData         = 0;
    ofn.lpfnHook          = NULL;
    ofn.lpTemplateName    = NULL;
}


BOOL FileOpenDlg (HWND hwnd, PTSTR pstrFileName, PTSTR pstrTitleName)
{
    FileInitialize(hwnd);
	ofn.hwndOwner         = hwnd;
    ofn.lpstrFile         = pstrFileName;
    ofn.lpstrFileTitle    = pstrTitleName;
    ofn.Flags             = OFN_HIDEREADONLY;
	
    return GetOpenFileName (&ofn);
}

BOOL FileSaveDlg (HWND hwnd, PTSTR pstrFileName, PTSTR pstrTitleName)
{
	FileInitialize(hwnd);
    ofn.hwndOwner         = hwnd;
    ofn.lpstrFile         = pstrFileName;
    ofn.lpstrFileTitle    = pstrTitleName;
    ofn.Flags             = OFN_HIDEREADONLY | OFN_OVERWRITEPROMPT;
	
    return GetSaveFileName (&ofn);
}



unsigned char* readImagePNG(BITMAPINFO& bitmapInfo, char* fileName)
{
	FILE* file;
	unsigned char ch, colorType ;
	unsigned int size;
	unsigned int imageSize=0, imageOffset=0, imageStartOffset=-1;
	unsigned char* imageData;

	unsigned char buf[8];
	memset(&bitmapInfo, 0, sizeof(BITMAPINFO));
	if ((file=fopen(fileName, "rb"))==NULL)
	{
		return NULL;
	}
	if (fread(buf, 1, 8, file)!=8)
	{
		return NULL;
	}
	if (memcmp(buf, PNG_Signature, 8)!=0)
	{
		return NULL;
	}
	//skip 8 bytes=length+type
	if (fseek(file, 8, SEEK_CUR)!=0)
	{
		return NULL;
	}

	if (fread(&size, 4, 1, file)!=1)
	{
		return NULL;
	}
	bitmapInfo.bmiHeader.biWidth=byteSwap(size);
	if (fread(&size, 4, 1, file)!=1)
	{
		return NULL;
	}
	bitmapInfo.bmiHeader.biHeight=byteSwap(size);

	if (fread(&ch, 1, 1, file)!=1)
	{
		return NULL;
	}

	bitmapInfo.bmiHeader.biBitCount=ch;

	if (fread(&colorType, 1, 1, file)!=1)
	{
		return NULL;
	}

	

	//4byte+CRC
	if (fseek(file, 7, SEEK_CUR)!=0)
	{
		return NULL;
	}

	while (true)
	{
		if ((size=readChunk(file, buf))==-1)
		{
			return NULL;
		}
		if (memcmp(buf, "IDAT", 4)==0)
		{
			imageSize+=size;
			if (imageStartOffset==-1)
			{
				imageStartOffset=ftell(file)-8;
			}
			
		}
		else
		{
			if (memcmp(buf, "IEND", 4)==0)
			{
				break;
			}
		}


		if (fseek(file, size+4, SEEK_CUR)!=0)
		{
			return NULL;
		}
	}

	imageData=new unsigned char[imageSize];
	if (fseek(file, imageStartOffset, SEEK_SET)!=0)
	{
		delete[]imageData;
		return NULL;
	}

	while (true)
	{
		if ((size=readChunk(file, buf))==-1)
		{
			delete[]imageData;
			return NULL;
		}
		if (memcmp(buf, "IDAT", 4)==0)
		{
			if (fread(imageData+imageOffset, size, 1, file)!=1)
			{
				delete[]imageData;
				return NULL;
			}
			imageOffset+=size;			
		}
		else
		{
			if (memcmp(buf, "IEND", 4)==0)
			{
				break;
			}
			else
			{
				delete[]imageData;
				return NULL;
			}
		}

		if (fseek(file, 4, SEEK_CUR)!=0)
		{
			return NULL;
		}
	}

	if (imageOffset!=imageSize)
	{
		printf("what is going on\n");
		delete[]imageData;
		return NULL;
	}

	bitmapInfo.bmiHeader.biSizeImage=imageSize;

	bitmapInfo.bmiHeader.biSize=sizeof(BITMAPINFOHEADER);
	
	bitmapInfo.bmiHeader.biPlanes=1;
	bitmapInfo.bmiHeader.biClrUsed=0;
	bitmapInfo.bmiHeader.biClrImportant=0;
	
	bitmapInfo.bmiHeader.biCompression=BI_PNG;
	fclose(file);
	return imageData;
}





float Hex::Hexagon[6][2]=
{
	{0.5, 0.134}, 
	{1.5, 0.134},
	{2.0, 1.0},
	{1.5, 1.866},
	{0.5, 1.866},
	{0, 1.0}
};

void Line::init(POINT p1, POINT p2)
{
	A=p1.y-p2.y;
	B=p2.x-p1.x;
	C=-(A*p1.x +B*p1.y);
}

void Line::init(POINT point, float w, float h)
{
	A=-h;
	B=w;
	fixLine(point);
}

void Line::fixLine(POINT point)
{
	C=-(A*point.x+B*point.y);
}

void Line::crossLine(Line& cross)
{
	cross.A=-B;
	cross.B=A;
	cross.C=C;
}

bool Line::evalX(int& x, int y)
{
	if (A==0)
	{
		return false;
	}
	x= -(B*y+C)/A;
	return true;
}

bool Line::evalY(int x, int& y)
{
	if (B==0)
	{
		return false;
	}
	y=-(A*x+C)/B;
	return true;
}

float pointDistance(POINT p1, POINT p2)
{
	return sqrt(p1.x*p2.x+p1.y*p2.y);
}

void Line::point2line(POINT point, float& distance)
{
	float dist;
	POINT p;
	int x, y;
	Line cross;
	crossLine(cross);

	fixLine(point);

	if (!evalY(point.x, y))
	{
		distance=abs(C/A-point.x);
		return;
	}

	if (!cross.evalX(x, y))
	{
		//this is impossible
		return;
	}

	p.x=x;
	p.y=y;

	dist=pointDistance(p, point);

	distance=(p.y-point.y)*(p.y-point.y)/dist;
}


void Hex::getHexPoints(POINT coord, POINT points[6])
{
	for (int i=0; i<6; i++)
	{
		points[i].x=size*Hexagon[i][0]+coord.x*offsetX;
		if (EVEN(coord.x))
		{
			points[i].y=size*Hexagon[i][1]+coord.y*offsetY;
		}
		else
		{
			points[i].y=size*Hexagon[i][1]+coord.y*offsetY+offsetY/2;
		}
	}
}

bool Hex::enumAroundHex(POINT hex, HEXENUM dir, POINT& result)
{
	//myHex.refresh();
	switch (dir)
	{
	case TopHex:
		result.x=hex.x;
		result.y=hex.y-1;
		break;
	case TopRightHex:
		result.x=hex.x+1;
		if (hex.x%2==0)
		{
			result.y=hex.y-1;
		}
		else
		{
			result.y=hex.y;
		}
		break;		
	case BottomRightHex:
		result.x=hex.x+1;
		if (hex.x%2==0)
		{
			result.y=hex.y;		
		}
		else
		{
			result.y=hex.y+1;
		}
		break;
	case BottomHex:
		result.y=hex.y+1;
		result.x=hex.x;
		break;		
	case BottomLeftHex:
		result.x=hex.x-1;
		if (hex.x%2==0)
		{
			result.y=hex.y;
		}
		else
		{
			result.y=hex.y+1;
		}
		break;
	case TopLeftHex:
		result.x=hex.x-1;
		if (hex.x%2==0)
		{
			result.y=hex.y-1;
		}
		else
		{
			result.y=hex.y;
		}
		break;
	}
	return result.x>0&&result.x<wStep&&result.y>0&&result.y<hStep;
}



void Hex::init(int w, int h, int theSize)
{
	width=w;
	height=h;
	size=theSize;
	wStep=width/(size*1.5);
	hStep=height/(size*1.732);
	leftFirst=size;
	topFirst=size;
	offsetX=1.5*size;
	offsetY=1.732*size;
	rectHalfSize=size*0.612;
}

void Hex::refresh()
{
	POINT point;
	RECT rect;
	GetClientRect(myWindow.hGlobalWnd, &rect);
	width=rect.right;
	height=rect.bottom;
	wStep=width/(size*1.5);
	hStep=height/(size*1.732);
	
	leftFirst=size;
	topFirst=size;
	
	offsetX=1.5*size;
	offsetY=1.732*size;
}

/*
bool Hex::coord2hex(POINT coord, POINT& hex)
{
	

}
*/

bool Hex::hex2center(POINT hex, POINT& center)
{
	if (hex.x<0||hex.y<0||hex.x==wStep||hex.y==hStep)
	{
		return false;
	}
	center.x=leftFirst+offsetX*hex.x;
	if (EVEN(hex.x))
	{
		center.y=topFirst+offsetY*hex.y;
	}
	else
	{
		center.y=topFirst+offsetY*hex.y+offsetY/2;
	}
	return true;
}



bool Hex::coord2hex(POINT coord, POINT& hex)
{
	int row, col;
	int w, h;
	int x, y;
	Line line;
	POINT point1, point2, origin;

	col=(coord.x-leftFirst)/offsetX;
	row=((coord.y-topFirst)*2+offsetY)/offsetY;

	//sprintf(buf, "startPoint[%d,%d]offset[%d,%d]size[%d]", leftFirst, topFirst,
	//	offsetX, offsetY, size);
	//MessageBox(NULL, buf, "caption", MB_OK);
	buf[0]='\0';

	sprintf(buf+strlen(buf), "coord[%d,%d][col=%d][row=%d]", coord.x, coord.y, col, row);

	if (row<0)
	{
		return false;
	}

	if (coord.x<leftFirst-size/2)
	{
		point1.x=leftFirst-size;
		point1.y=topFirst+(row/2)*offsetY;

		point2.x=leftFirst-size/2;
		if (EVEN(row))
		{
			point2.y=topFirst+(row/2)*offsetY-offsetY/2;
		}
		else
		{
			point2.y=topFirst+(row/2)*offsetY+offsetY/2;
		}

		line.init(point1, point2);
		line.evalY(coord.x, y);
		if (EVEN(row) && coord.y>y|| ODD(row)&& coord.y<y)
		{	
				hex.x=0;
				hex.y=row/2;
				return true;
		}
		else
		{
			return false;
		}
		sprintf(buf+strlen(buf), "{0}{coord.y:y=%d,%d", coord.y, y);
	}
		
	//important! (row/2) is calculated as integer division
	point1.y=topFirst+(row/2)*offsetY;

	//h=myHex.offsetY;

	if (EVEN(col))
	{
		point1.x=leftFirst+col*offsetX+size;
		point2.x=leftFirst+col*offsetX+size/2;
	}
	else
	{
		point1.x=leftFirst+(col+1)*offsetX-size;
		point2.x=leftFirst+(col+1)*offsetX-size/2;
	}

	if (EVEN(col)&& EVEN(row) || ODD(col)&& EVEN(row))
	{
		//w=-myHex.size;
		point2.y=topFirst-offsetY/2 + (row/2)*offsetY;
	}
	if (EVEN(col) && ODD(row) || ODD(col)&& ODD(row))
	{
		//w=myHex.size;
		point2.y=topFirst+offsetY/2+ (row/2)*offsetY;
	}

	//line.init(point, w, h);
	line.init(point1, point2);
	line.evalY(coord.x, y);


	if (EVEN(col) && EVEN(row) && coord.y>y)
	{	
		hex.x=col;
		hex.y=row/2;
		sprintf(buf+strlen(buf), "{1}{coord.y>y=%d>%d[%f,%f,%f]point1[%d,%d]point2[%d,%d]", 
			coord.y, y, line.A, line.B, line.C, point1.x, point1.y, point2.x, point2.y);
		return true;
	}

	if (EVEN(col) && ODD(row)&& coord.y<y)
	{
		hex.x=col;
		hex.y=row/2;
		sprintf(buf+strlen(buf), "{8}{coord.y<y=%d<%d[%f,%f,%f]point1[%d,%d]point2[%d,%d]", 
			coord.y, y, line.A, line.B, line.C, point1.x, point1.y, point2.x, point2.y);
		return true;
	}

	if (ODD(col) && EVEN(row) && coord.y>y)
	{
		hex.x=col+1;
		hex.y=row/2;
		sprintf(buf+strlen(buf), "{2}{coord.y>y=%d>%d[%f,%f,%f]point1[%d,%d]point2[%d,%d]", 
			coord.y, y, line.A, line.B, line.C, point1.x, point1.y, point2.x, point2.y);
		return true;
	}

	if (ODD(col)&& ODD(row)&&coord.y<y)
	{
		hex.x=col+1;
		hex.y=row/2;
		sprintf(buf+strlen(buf), "{3}{coord.y>y=%d>%d[%f,%f,%f]point1[%d,%d]point2[%d,%d]", 
			coord.y, y, line.A, line.B, line.C, point1.x, point1.y, point2.x, point2.y);
		return true;
	}

	if (EVEN(col) && EVEN(row) && coord.y<y)
	{
		hex.x=col+1;
		hex.y=row/2-1;
		sprintf(buf+strlen(buf), "{4}{coord.y>y=%d>%d[%f,%f,%f]point1[%d,%d]point2[%d,%d]", 
			coord.y, y, line.A, line.B, line.C, point1.x, point1.y, point2.x, point2.y);
		if (hex.y<0)
		{
			return false;
		}
		return true;
	}

	if (ODD(col)&& EVEN(row)&& coord.y<y)
	{
		hex.x=col;
		hex.y=row/2-1;
		sprintf(buf+strlen(buf), "{5}{coord.y>y=%d>%d[%f,%f,%f]point1[%d,%d]point2[%d,%d]", 
			coord.y, y, line.A, line.B, line.C, point1.x, point1.y, point2.x, point2.y);
		if (hex.y<0)
		{
			return false;
		}
		return true;
	}


	if (EVEN(col) && ODD(row) &&coord.y>y)
	{
		hex.x=col+1;
		hex.y=row/2;
		sprintf(buf+strlen(buf), "{6}{coord.y>y=%d>%d[%f,%f,%f]point1[%d,%d]point2[%d,%d]", 
			coord.y, y, line.A, line.B, line.C, point1.x, point1.y, point2.x, point2.y);
		return true;
	}
	if (ODD(col)&& ODD(row)&&coord.y>y)
	{
		hex.x=col;
		hex.y=row/2;
		sprintf(buf+strlen(buf), "{7}{coord.y>y=%d>%d[%f,%f,%f]point1[%d,%d]point2[%d,%d]", 
			coord.y, y, line.A, line.B, line.C, point1.x, point1.y, point2.x, point2.y);
		return true;
	}

	return false;
}


bool Hex::hex2rect(POINT hex, RECT& rect)
{
	POINT origin;
	if (hex.x<0||hex.y<0||hex.x>=wStep||hex.y>=hStep)
	{
		return false;
	}
	//clientAreaOrg(origin);

	rect.left=leftFirst+offsetX*hex.x-rectHalfSize;
	if (EVEN(hex.x))
	{
		rect.top=topFirst+offsetY*hex.y-rectHalfSize;
	}
	else
	{
		rect.top=topFirst+offsetY*hex.y+ offsetY/2-rectHalfSize;
	}
	rect.right=rect.left+rectHalfSize*2;
	rect.bottom=rect.top+rectHalfSize*2;

	return true;
}



HBITMAP convertBM4Blending(HDC hdc, HBITMAP hBitmap)
{
	int result;
	BITMAP bitmap;
	unsigned char* buf;
	HBITMAP hResultBitmap;
	BITMAPINFO bitmapInfo;
	GetObject(hBitmap, sizeof(BITMAP), &bitmap);
	bitmapInfo.bmiHeader.biSize=sizeof(BITMAPINFOHEADER);
	bitmapInfo.bmiHeader.biWidth=bitmap.bmWidth;
	bitmapInfo.bmiHeader.biHeight=bitmap.bmHeight;
	bitmapInfo.bmiHeader.biPlanes=1;
	bitmapInfo.bmiHeader.biBitCount=32;
	bitmapInfo.bmiHeader.biCompression=BI_RGB;
	bitmapInfo.bmiHeader.biSizeImage=0;
	bitmapInfo.bmiHeader.biXPelsPerMeter=0;
	bitmapInfo.bmiHeader.biYPelsPerMeter=0;
	bitmapInfo.bmiHeader.biClrUsed=0;
	bitmapInfo.bmiHeader.biClrImportant=0;

	buf=new unsigned char[bitmap.bmWidth*bitmap.bmHeight*4];

	result=GetDIBits(hdc, hBitmap, 0, bitmap.bmHeight, buf, &bitmapInfo, DIB_RGB_COLORS);

	//hResultBitmap=CreateDIBitmap(hdc, &bitmapInfo.bmiHeader, CBM_INIT, buf, &bitmapInfo, 
	//	DIB_RGB_COLORS);
	hResultBitmap=CreateBitmap(bitmap.bmWidth, -bitmap.bmHeight, 1, 32, buf);

	return hResultBitmap;
}
	



void drawSun(HWND hwnd, POINT coord, int radius, int lightLength, int lightNumber)
{
	POINT origin;
	HDC hdc;
	HPEN hPen;
	HBRUSH hBrush;
	XFORM xForm, xForm1, xForm2, xForm3;

	hPen=CreatePen(PS_SOLID, 2, RGB(255, 0, 0));
	
	hdc=GetWindowDC(hwnd);

	clientAreaOrg(origin, hwnd);

	coord.x+=origin.x;
	coord.y+=origin.y;

	hPen=(HPEN)SelectObject(hdc, hPen);
	translate(coord.x, coord.y, xForm1);

	SetGraphicsMode(hdc, GM_ADVANCED);
	identity(xForm3);
	hBrush=CreateSolidBrush(RGB(255, 0, 0));

	hBrush=(HBRUSH)SelectObject(hdc, hBrush);

	Ellipse(hdc, coord.x-radius, coord.y-radius, coord.x+radius,coord.y+radius); 

	for (int i=0; i<360; i+=360/lightNumber)
	{			
		rotate(i, xForm2);
		CombineTransform(&xForm, &xForm2, &xForm1);

		SetWorldTransform(hdc, &xForm3);
		
		MoveToEx(hdc, radius+10, 0, NULL);

		SetWorldTransform(hdc, &xForm);
				
		LineTo(hdc, radius+lightLength, 0);
	}
	SetWorldTransform(hdc, &xForm3);
	hPen=(HPEN)SelectObject(hdc, hPen);
	hBrush=(HBRUSH)SelectObject(hdc, hBrush);
	DeleteObject(hBrush);
	DeleteObject(hPen);
	ReleaseDC(hwnd, hdc);
}


void doPrintDisplay(DISPLAY_DEVICE& dispDev, char* msg)
{
	sprintf(msg, "deviceName=%s", dispDev.DeviceName);

	sprintf(msg+strlen(msg), "deviceName=%s[flag]", dispDev.DeviceString);

	switch (dispDev.StateFlags)
	{
	case DISPLAY_DEVICE_ATTACHED_TO_DESKTOP:
		sprintf(msg+strlen(msg), "DISPLAY_DEVICE_ATTACHED_TO_DESKTOP");
		break;
	case DISPLAY_DEVICE_MIRRORING_DRIVER:
		sprintf(msg+strlen(msg), "DISPLAY_DEVICE_MIRRORING_DRIVER");

		break;
	//case DISPLAY_DEVICE_MODESPRUNED:
	//	sprintf(msg+strlen(msg), "DISPLAY_DEVICE_MODESPRUNED");

	//	break;

	//case DISPLAY_DEVICE_REMOVABLE:
	//	sprintf(msg+strlen(msg), "DISPLAY_DEVICE_REMOVABLE");

	//	break;
	case DISPLAY_DEVICE_VGA_COMPATIBLE:
		sprintf(msg+strlen(msg), "DISPLAY_DEVICE_VGA_COMPATIBLE");

		break;
	}
}


void checkDisplay(PrintMessageMethod printMethod)
{
	DISPLAY_DEVICE dispDev;
	char msg[256];
	dispDev.cb=sizeof(DISPLAY_DEVICE);
	int i=0;
	while (EnumDisplayDevices(NULL, i, &dispDev, 0))
	{
		doPrintDisplay(dispDev, msg);
		printMethod(msg);
		i++;
	}
}

HRGN HexRgn::getRgn(POINT coord)
{
	POINT hexCoord;
	hex.coord2hex(coord, hexCoord);
	return hexRgn[hexCoord.y*hex.wStep+hexCoord.x];
}


void HexRgn::init(HWND hwnd, int size)
{	
	POINT points[6], coord;
	RECT rect;
	GetClientRect(hwnd, &rect);

	hex.init(rect.right, rect.bottom, size);
		
	hexRgn=new HRGN[hex.wStep*hex.hStep];
	winRgn=CreateRectRgn(0, 0, 0, 0);
	for (int i=0; i<hex.hStep; i++)
	{
		for (int j=0; j<hex.wStep; j++)
		{		
			coord.x=j;
			coord.y=i;
			hex.getHexPoints(coord, points);
		
			hexRgn[i*hex.wStep+j]=CreatePolygonRgn(points, 6, WINDING);
			CombineRgn(winRgn, winRgn, hexRgn[i*hex.wStep+j], RGN_OR);						
		}
	}
}

	
//to be continued
void brightenRgn(HDC hdc, HRGN hRgn, float alpha)
{
	RECT rect;
	GetRgnBox(hRgn, &rect);
}